/* Program to read-modify-write video modes for higher frequency */
/* Requires VESA VBE 3 */

#include <stdio.h>  /* sprintf, printf */
#include <dos.h>    /* MK_FP, _psp, int86, union REGS */ 
#include <string.h> /* memcpy */
#include <stdlib.h> /* atoi, atol */

#define MIO 1000000UL    /* guess what. One million. */

#define MAXFRAMES 140UL  /* maximum framerate in Hz */
#define MAXeFRAMES 155UL /* higher maximum framerate in Hz during 1st estimate */

#define GUESSXREF 120UL  /* percentage factor xres -> htotal for first guess */

#undef CRTCREGDUMP       /* define to show CRTC register dump after mode change */
#undef CRTCFIX           /* define to enforce CRTC hsyncpos <= hvisible   */
#define TRYVARIANTS 1    /* define to enable interactive modifications... */
#define FONTFOO 1        /* define to treat 9x?? font modes as 8x?? ones  */

#define TEXTCLOCK (55UL*MIO) /* maximum pixel clock for text modes */
  /* All cards which can do 132x60 at 70 Hz can do 35 Mhz. Basically all  */
  /* SVGA cards should reach 45 MHz. Do not try above 70 MHz, even if the */
  /* card is fast in GRAPHICS mode: Text modes cannot use fast RAM bursts */
  /* so you will get display errors even on modern cards at some point... */

/* #define SYNCPOLARITY 1 */ /* define to have xor sync polarity for 350 / 400 lines */

#define DMODE 1        /* define for debug mode */ 
/* #define DMODE2 1 */ /* define for more verbose debug mode */ 

#ifdef DMODE2
#define DMODE 1     /* verbose debug mode implies normal debug mode */
#endif

#define word unsigned int
#define byte unsigned char
#define dword unsigned long int

typedef struct CRTC_params
{
  word htotal, hsync1, hsync2, vtotal, vsync1, vsync2;
  byte crtcflags; /* bit 0 doublescan, 1 interlace, 2 negvsync 3 neghsync */
  dword clock;
  word vrefresh; /* must be 100 * clock / (htotal * vtotal) */
  byte rsvd[40];
} CRTCPARAMS;

typedef struct VESA_mode
{
  word attrib;
  byte winattrib1, winattrib2;
  word wingranularity, winsize, winseg1, winseg2;
  void far * winfunc;
  word bpline, xres, yres; /* res in char units if text mode */
  byte xchar, ychar, planes, bpp, banks, model, pages, vbe3flag;
  /* */
  byte rmask, rshift, gmask, gshift, bmask, bshift, amask, ashift, colorattrib;
  /* VBE2+: */
  dword lfbaddr;
  char far * offscreenptr;
  word offscreenram;
  /* VBE3+: */
  word lfbbpline;
  byte maxpagebanked, maxpagelfb;
  byte lfbrmask, lfbrshift, lfbgmask, lfbgshift, lfbbmask, lfbbshift;
  byte lfbamask, lfbashift;
  dword maxclock; /* <-- */
  word modenumber; /* *** ADDED for VBE3fast *** */
} VESAMODE;

union REGS r;

int Xinp(int port);
int Xinp(int port) /* slow INPUT */
{
  int p = port+1;
  int x = inportb(p-1);
  /*  printf("<%x>:%x ", p-1, x); */
  return x;
}


/* 10.4f0b.bl=0.dx=mode.ecx=requestedclock -> ecx possibleclock! */
/* Heavy Turbo C 2.01 abuse ahead! Turbo C has no 386 awareness. */
unsigned long realclock(unsigned long desiredclock, unsigned int mode);

unsigned long realclock(unsigned long desiredclock, unsigned int mode)
{
  register unsigned long actualclock;
  _AX = 0x4f0b;
  _BX = 0;
  _CX = desiredclock >> 16;  /* high part first  */
  __emit__(0x66,0xC1,0xE1,0x10); /* shl ecx,0x10 */
  _CX = (unsigned int) desiredclock;
  _DX = mode;                /* because clock choice can depend on mode */
  __emit__(0xcd,0x10);       /* int86(0x10, ...) */
  if (_AX != 0x004f)         /* error detected?  */
    return 0;                /* (returned value is in DX AX) */
  actualclock = _CX;         /* low part first   */
  __emit__(0x66,0xC1,0xE9,0x10); /* shr ecx,0x10 */
  return actualclock | ((unsigned long)_CX << 16);
}

/* cols and lines can use either chars or pixels as units */
/* mask1 is ANDed with the mode attributes and result must be mask2 */
int set_vesa (int cols, int lines, int mask1, int mask2, VESAMODE * vmode);

int set_vesa(int cols, int lines, int mask1, int mask2, VESAMODE * vmode)
{
  unsigned char vesabuffer[256];
  unsigned int vesamodes[150]; /* we have to copy the list! */
  VESAMODE * myvmode = (VESAMODE *)&vesabuffer[0];
  unsigned int i, j, xcols, xlines, minbpp;
  unsigned long li;
  char far * str;
  unsigned int far * model;
  struct SREGS sregs;

  r.h.ah = 0x12; /* ah=0x12, bl=0x10: get EGA info */
  r.x.bx = 0xff10; /* init bh to -1 */
  r.x.cx = 0xffff; /* init cx to -1 */
  int86(0x10, &r, &r);
  if ( (r.x.cx == 0xffff) || (r.h.bh == 0xff) )
    return 1; /* not even EGA */
  r.x.ax = 0x1a00; /* get VGA display combination - VGA install check */
  int86(0x10, &r, &r);
  if (r.h.al != 0x1a)
    return 1; /* not even VGA */

  for (i = 0; i < sizeof(vesabuffer); i++)
    vesabuffer[i] = ' ';
  sregs.es = FP_SEG(&vesabuffer[0]);
  r.x.di = FP_OFF(&vesabuffer[0]);
  r.x.ax = 0x4f00; /* get VESA version and mode list */
  int86x(0x10, &r, &r, &sregs); /* call VESA version check */
  if ( (r.x.ax != 0x004f) || (vesabuffer[0]!='V') || (vesabuffer[1]!='E') ||
       (vesabuffer[2]!='S') || (vesabuffer[3]!='A') ||
     ( ((vesabuffer[5] == 1) && (vesabuffer[4] < 2)) || (vesabuffer[5] < 1) )
     ) {
    printf("No VESA, but %d x %d mode not supported on plain VGA yet.\r\n",
      cols, lines);
    if (r.x.ax == 0x004f) printf("VESA 1.2 or newer required.\r\n");
    return 1;
  }
  if (vesabuffer[5] < 3) {
    printf("* Simulating: Real clock tuning is only available if at least VESA VBE 3.0!\r\n");
    /* could loosen requirements to modes in this case */
  }

  /* some BIOSes return the mode list *in* the vesabuffer! */
  /* so we have to copy it to protect it from overwriting. */
  model = MK_FP( peek(sregs.es, FP_OFF(&vesabuffer[0x10])),
    peek(sregs.es, FP_OFF(&vesabuffer[0x0e])) );
  i = 0;
  while ( (*model != 0xffff) && (i<149) ) {
    vesamodes[i] = *model;
    if (vesamodes[i]) i++; /* do not store the value 0 */
    model++;
  }
  for (j = 1; (j <= 0x13) && (i<149); j++)
    vesamodes[i++] = j; /* add (most) standard VGA modes */
  vesamodes[i] = 0; /* we use 0, not -1, as end marker */

  r.x.ax = 0x4f03; /* get current VESA mode */
  int86(0x10, &r, &r); /* query current VESA mode */
  /* current mode is now in BX - high 7 bits are for flags: */
  /* 0x8000 did not clear screen 0x4000 linear framebuffer */

#ifdef DMODE
  printf("VESA %hu.%hu '", vesabuffer[5], vesabuffer[4]);
  li = vesabuffer[0x13];
  li <<= 8;
  li |= vesabuffer[0x12];
  str = MK_FP( peek(sregs.es, FP_OFF(&vesabuffer[8])),
    peek(sregs.es, FP_OFF(&vesabuffer[6])) );
  while (*str) {
    printf("%c",*str);
    str++;
  }
  printf("', mode 0x%04x, %luk for BIOS, %u modes.\r\n",
    r.x.bx & 0x01ff, li * 64, i);
#endif

  r.x.cx = r.x.bx & 0x01ff;
  r.x.ax = 0x4f01; /* get info about VESA mode */
  sregs.es = FP_SEG(myvmode);
  r.x.di = FP_OFF(myvmode);
  int86x(0x10, &r, &r, &sregs);

  if ( (r.x.ax == 0x004f) && ((myvmode->attrib & mask1) == mask2) ) {
       /* info valid and BIOS supported text mode: 0x17 0x06 */
    if ( ( (!cols) ||  (myvmode->xres == cols)  ) &&
         ( (!lines) || (myvmode->yres == lines) ) ) {
#ifdef DMODE2
      printf("Current mode 0x%04x already has requested size. Attrib 0x%04x\r\n",
        r.x.cx, myvmode->attrib);
#endif
#ifdef FONTFOO
      myvmode->xchar = (myvmode->xchar == 9) ? 8 : myvmode->xchar;
#endif
      myvmode->modenumber = r.x.cx;
      memcpy(vmode, myvmode, sizeof(VESAMODE));
      return 0;
    } /* matching mode */
    /* first, try to find a mode which preserves unspecified variables */
    xcols = (cols) ? cols : myvmode->xres;
    xlines = (lines) ? lines : myvmode->yres;
  } /* valid mode */

  i = 0;  /* start with low-numbered modes */
  li = 0; /* try good fit first */
  minbpp = 8; /* try 256 color modes first */
  while (1) {
    while (vesamodes[i]) {
      r.x.cx = vesamodes[i] & 0x01ff;
      r.x.ax = 0x4f01; /* get info about VESA mode */
      sregs.es = FP_SEG(myvmode);
      r.x.di = FP_OFF(myvmode);
      int86x(0x10, &r, &r, &sregs);
      if ( (r.x.ax == 0x004f) && ((myvmode->attrib & mask1) == mask2) ) {
           /* info valid and BIOS supported reachable text mode 0x16 0x06 */
           /* bits: 0 reachable??, 1 cols/lines known, 2 BIOS supported */
           /* 3 color, 4 graphical, 5 non-VGA (VBE 2.0+) ... 12 dual    */
        if ( ( (!xcols) || (myvmode->xres == xcols)  ) &&
             ( (!xlines) || (myvmode->yres == xlines) ) &&
             (myvmode->bpp >= minbpp) /* no too simple modes please */ ) {

          pokeb(0x40,0x89, peekb(0x40,0x89) |
            ( ((myvmode->yres >= 200) && (myvmode->yres <= 350))
              ? 0 : 0x10 ) );
            /* bit 4 is 0 for EGA 1, but for VGA (before mode switch) */

          r.x.ax = 0x4f02; /* set VESA mode */
          r.x.bx = vesamodes[i] & 0x01ff;
          int86(0x10, &r, &r);
          printf("VESA mode 0x%04x: %ux%u (%hu bpp), %hux%hu font, attrib 0x%04x.\r\n",
            vesamodes[i], myvmode->xres, myvmode->yres, myvmode->bpp,
            myvmode->xchar, myvmode->ychar, myvmode->attrib);
          /* could also check: 0x1b is 0 for text mode memory layout, */
          /*   0x13 and 0x14 are high bytes, 0x10 is memory line len. */
          myvmode->modenumber = vesamodes[i] & 0x1ff;
#ifdef FONTFOO
         myvmode->xchar = (myvmode->xchar == 9) ? 8 : myvmode->xchar;
#endif
          memcpy(vmode, myvmode, sizeof(VESAMODE));
          return 0; /* SUCCESS! */
        } /* matching mode */
#ifdef DMODE2
        else printf("  [Not %u x %u (%hu bpp) mode 0x%04x]",
          myvmode->xres, myvmode->yres, myvmode->bpp, vesamodes[i]);
#endif        
      } /* valid mode: BIOS text mode */
      i++; /* try next mode */
    } /* while */

    if (li) {
#ifdef DMODE2
      printf("\r\nSorry, no such VESA mode found.\r\n");
#endif
      return 1; /* not even a sloppy match worked */
    }

    if (minbpp > 2) {
      minbpp >>= 1; /* try to find 256 -> 16 -> 4 color modes */
      i = 0; /* start searching again */
      continue;
    }

    if ((!cols) || (!lines)) { /* only 1 specified? */
#ifdef DMODE2
      printf("\r\nNot possible to change only %s. Allowing %s to change, too.\r\n",
        (cols) ? "columns" : "lines", (cols) ? "lines" : "columns");
#endif
      minbpp = 8; /* try at least 256 colors again */
      /* if nothing found yet, allow any value for unspecified variables */
      xcols = cols;
      xlines = lines;
      i = 0; /* start searching again */
      li = 1; /* flag: already searching for sloppy fit */
    } else { /* no sloppyness possible */
#ifdef DMODE2
      printf("\r\nSorry, no such VESA mode. Try setting ONLY cols or ONLY lines.\r\n");
#endif
      return 1;
    } /* relax if sloppy, give up else */

  } /* loop 1st good fit, 2nd sloppy fit */

  /* return 1; - never reached */ /* failed */
} /* set_vesa */




/* *** MAIN() of this VESA VBE3 custom CRTC / refresh rate program *** */
int main(int argc, char ** argv)
{
  int xres, yres, khz, i;
  unsigned int dscan, egamode = 0; /* flags */
  unsigned long xref;
  struct SREGS sregs;
  struct VESA_mode mode;
  struct CRTC_params cparm;

  if (argc < 4) { /* counting the 0th argument as SELF */
    printf("Arguments must be 3 numbers which represent the X and Y resolution\r\n");
    printf("of a graphics or text mode in pixels or characters and the desirde\r\n");
    printf("hsync rate in kHz. The vsync rate is kept below %lu Hz.\r\n",
      MAXFRAMES);
#ifdef TRYVARIANTS
    printf("If any 4th argument is given, interactive mode editing is enabled.\r\n");
#endif
  }

  xres = atoi(argv[1]);
  yres = atoi(argv[2]);

  dscan = ((yres >= 200) && (yres <= 350 /* 300 ??? */)) ? 0x100 : 0;
    /* only accept doublescanned modes for 320x200 / 320x240 / ... */
  if ((xres > 320) && dscan) {
    dscan = 0; /* remove doublescan property for 640x200 / 640x350 !?!? */
    egamode = 1; /* flag for EGA modes 640x200 / 640x200 */
  }

  khz = atoi(argv[3]);

  if ( (xres < 40) || (yres < 15) || (xres > 1600) || (yres > 1200) ||
      (khz < 30) || (khz > 100) ) {
    printf("Current version only allows 40..1600 x 15..1200 at 30..100 kHz\r\n");
    return 1;
  }

#if 0
  r.x.ax = 0003; /* 80x25 default mode and clear screen */
  int86(0x10,&r,&r);
#endif

  xref = (unsigned)khz * 1000UL;

  i = set_vesa(xres, yres, 0x06 | dscan, 0x06 | dscan, &mode);
    /* 0x17 0x07: reachable text mode */
    /* 0x06 0x06: any mode with TTY functions */

  if (i) {
    printf("No mode found for %d x %d resolution.\r\n", xres, yres);
    return 1;
  }

  if (!(mode.attrib & 1)) printf(" [Unreachable!?] ");
  /* & 2 has to be nonzero anyway */
  printf("%s ", (mode.attrib & 0x04) ? "BIOS TTY supported" : "Non-TTY-BIOS");
  printf("%s ", (mode.attrib & 0x08) ? "color" : "monochrome");
  printf("%s mode\r\n", (mode.attrib & 0x10) ? "graphics" : "text");
  if (mode.attrib & 0x020) printf("VGA compatible");
  /* & 0x40 nonzero if VGA memory windows available */
  if (mode.attrib & 0x080) printf(" [LinearFrameBuffer-able]");
  if (mode.attrib & 0x100) printf(" [DoubleScan-able]");
  if (mode.attrib & 0x200) printf(" [Interlace-able]");
  if (mode.attrib & 0x3a0) printf("\r\n");
  /* & 0x400 nonzero if hardware triplebuffering available */
  /* & 0x800 nonzero if hardware stereo display available  */
  /* & 0x1000 nonzero if dual display start address available  */

  if (!(mode.attrib & 0x10)) {
#ifdef DMODE2
    printf("  Mode is %d x %d text -> ", mode.xres, mode.yres);
#endif
    mode.xres *= mode.xchar;
    mode.yres *= mode.ychar;
#ifdef DMODE2
    printf("%d x %d pixels\r\n", mode.xres, mode.yres);
#endif
    /* modes with 9 pixel wide chars reach higher overall clock at the */
    /* same "char slice" clock (chars per line * pixel lines per sec). */
  }

#ifdef DMODE2
  printf("  Mode 0x%04x is %d x %d pixels, max %lu MHz pixel clock.\r\n",
    mode.modenumber, mode.xres, mode.yres, mode.maxclock/MIO);
#endif

  { /* variable scope: CRTC params */
    unsigned long yref1, clock1, clock2, i;
    unsigned long hpulse, vpulse, hpos1, hpos2, vpos1, vpos2;

    if (mode.yres <= 300) {
      mode.yres *= 2; /* treat double-scanned modes like normal modes */
                      /* with the double vertical resolution... !!!   */
      dscan = 0x100;
      printf("  Using doublescan mode.\r\n");
    }

    if ( (egamode) && (xref != ((mode.yres == 350) ? 32000UL : 30000UL)) ) {
      printf("EGA 640x350 and 640x200 need 31.5 kHz and 30-31.5 kHz respectively.\r\n");
      /* actually 640x200 is normally 70 Hz 31.5 kHz and we use 70 Hz and 30 kHz   */
      /* ... and  640x350 is normally 70 Hz 31.5 kHz and we use 85 Hz and 31.9 kHz */
      /* -> looks like we use too small EGA vertical blanking in general? */
      /* However, image errors at other freq is: top / left of image very clipped! */
      if ( argc <= 4 ) /* unless interactive */
        return 1;
      else {
        printf(">>>>.>>>>.>>>>.>>>>.>>>>.>>>>.>>>>. Left 25-50%%  cannot  be displayed at\r\n");
        printf(">>>>.>>>>.>>>>.>>>>.>>>>.>>>>.>>>>. nonstandard freq for EGA 640x???.\r\n");
        /* no idea why, but nothing helps (hpos1=0, small hpos2, big hpulse). */
      }
    }

    clock1 = (GUESSXREF*mode.xres/100UL) * xref; /* take a guess based on xref */
      /* assume htotal = 125% xres? Or better more? */

    if ((!mode.maxclock) || (mode.maxclock > TEXTCLOCK)) {
      if (!(mode.attrib & 0x10)) { /* text mode */
        if (mode.maxclock > TEXTCLOCK)
          printf("* Text mode -> max. %lu MHz, not %lu MHz.\r\n",
            TEXTCLOCK/MIO, mode.maxclock/MIO);
        mode.maxclock = TEXTCLOCK; /* text mode: all 132x60 cards can do 45 MHz */
        /* even 70 MHz should be safe on most cards although text modes are "out" */
      } else {
        if (!mode.maxclock) {
          printf("* Simulating: DAC speed unknown, using 110 MHz\r\n");
          mode.maxclock = 110UL * MIO; /* invent 110 MHz DAC */
        } else {
          if (mode.maxclock > (300UL*MIO)) {
            mode.maxclock = 300UL*MIO; /* enough for 1600x1200 156 Hz... */
            printf("* Too high / not plausible DAC speed, using 300 MHz.\r\n");
          }
        }
      } /* graphics mode */
    } /* DAC speed sanity checks */

    if (clock1 > mode.maxclock) {
      printf("Would need %lu MHz bandwidth but only %lu possible.\r\n",
        clock1/MIO, mode.maxclock/MIO);
      return 1;
    }

    yref1 = xref / (105UL*mode.yres/100UL);
        /* using real xref, not estimate: we want to hit the   exact    */
        /* desired hrefresh rate later and vrefresh depends on hrefresh */
        /* we can have much more than 5% blank lines, so let us be less */
        /* pessimistic about ending up with too high vrefresh rate...   */

#ifdef DMODE2
    { /* xref1 estimate */
      unsigned long xref1;
      xref1 = clock1 / (GUESSXREF*mode.xres/100UL);
      if (egamode)
        xref1 = clock1 / (GUESSXREF*mode.xres/110UL);
          /* avoid big guesses for EGA, blanking should stay stort */
          /* assume htotal = 125% xres? Or better more? */
      printf("High estimate hrefresh %lu kHz vrefresh %lu Hz at %lu MHz.\r\n",
        xref1/1000UL, yref1, clock1/MIO);
    } /* xref1 estimate */
#endif

    clock2 = clock1;
    clock1 = realclock(clock1, mode.modenumber); /* find next possible hardware clock */

#ifdef DMODE2
    printf("Actually used clock will be %lu kHz (not %lu kHz).\r\n",
      clock1/1000UL, clock2/1000UL);
#endif

    if (clock1 < (10UL*MIO)) { /* less than 10 MHz is too strange */
      printf("* Real clock %lu kHz? Will not work. Simulating.\r\n",
        clock1/1000UL);
      clock1 = clock2;            /* pretend perfect clock */
      clock2 = 0;                 /* mark simulation */
    }

    /* CRTs are often happiest with 1-4ysec. Unit: pixels. EGA used 4ysec. */
    /* older CRTs want 2ysec or even 3ysec. TFT accept anything 1-8ysec.   */
    /* *** If the pulse is too short, image will be too wide. *** */
    hpulse = clock1 / 480000UL; /* 2.1ysec */
    /* EGA/VGA uses roughly 1.6ysec. SVGATextMode suggests "anything > 0". */
    /* *** If this value is too small, right margin will be too small ***  */
    hpos1 = clock1 / 2500000UL; /* 0.4ysec */ /* (hpulse++ -> hpos1-- ok)  */
    /* post-sync should be e.g. 1.2-3.8ysec. Using 1.3ysec for now...?     */
    /* EGA only has 1ysec? Nice minimum: pre 0.5 pulse 1.2 post 1.3 sum 4  */
    /* *** > 1.5ysec / < 1.3ysec -> too much / too small left margin. ***  */
    hpos2 = clock1 / 720000UL; /* 1.4ysec */
    /* is getting a sum of >= 4ysec important?  */

    { /* rounding */
      unsigned int xgranularity;

      xgranularity = 4; /* for graphics modes */
      if ((mode.xchar >= 8) && (xgranularity <= 9))
        xgranularity = mode.xchar; /* of X timing values */

      hpulse += xgranularity - 1;
      hpos1 += xgranularity - 1;
      hpos2 += xgranularity - 1;

      hpulse = (hpulse / xgranularity) * xgranularity;
      hpos1 = (hpos1 / xgranularity) * xgranularity;
      hpos2 = (hpos2 / xgranularity) * xgranularity;

      if (xgranularity == 4)
        mode.xchar = 4; /* unelegant way to store xgranularity but works */
    } /* rounding */

    i = clock1 / (mode.xres + hpos1 + hpulse + hpos2);
    if ((i/97UL) < (xref/100UL)) { /* considerably too LOW hrefresh clock? */
#ifdef DMODE2
      printf("Pumping dot clock a bit, hrefresh was too low.");
#endif
      if (!clock2) {
        clock1 = (clock1/100UL) * 108UL; /* give 8% more for simulation */
      } else {
        clock1 = realclock((clock1/100UL) * 108UL /* try 8% more here */,
          mode.modenumber); /* find next possible hardware clock */
      }
      i = clock1 / (mode.xres + hpos1 + hpulse + hpos2);
#ifdef DMODE2
      printf(" %lu MHz now.\r\n", clock1/MIO);
#endif
    }

#ifdef DMODE2
    if (i > xref) /* maximum hrefresh > desired hrefresh? */
      printf("Maximum possible hrefresh clock: %lu (%u + %lu + %lu + %lu)\r\n", i,
        mode.xres, hpos1, hpulse, hpos2);
#endif

    while (i > xref) { /* inflate blanking area to lower hrefresh rate */
      hpulse += mode.xchar; /* one "granularity unit" more */
      i = clock1 / (mode.xres + hpos1 + hpulse + hpos2);
        /* more hpos1 moves image LEFT and increases crtc[4] and crtc[5] ? */
        /* more hpulse / less hpos2 (?) has the same effect ? */
        /* less hpos1 brings the image closer to the right border. */
    };

    printf("Found hrefresh %lu.%03lu kHz at %u + %lu + %lu + %lu.\r\n",
      i/1000UL, i%1000UL, mode.xres, hpos1, hpulse, hpos2);

    if (i < 30000UL) {
      printf("Horizontal frequency of < 30 kHz is not acceptable.\r\n");
      return 1;
    }
    if (mode.xres < ((hpos1 + hpulse + hpos2)*2)) {
      printf("Too much of each line is blanking. Probably too high pixel clock.\r\n");
      return 1;
    }

    /* now that we know that we have i lines per second, we can use that */
    /* to create an 50-300ysec (TFT: 40-500ysec?) vsync pulse. Two lines */
    /* (or 1..6) are popular. But a bit low for high frequency modes...? */
    vpulse = (i / 15625UL) + 1; /* 64ysec (min 45ysec recommended) */
    if (egamode)
      vpulse = (i / 22222UL) + 1; /* 45ysec: not enough top margin if >? */

    /* 100ysec or even less seem to be okay but 300ysec are popular... */
    vpos1 = (i / 3333UL) + 1; /* 300ysec */
    if (egamode)
      vpos1 = (i / 2500UL) + 1; /* 400ysec */

    /* back porch of vertical blanking is long for CRTs... 600ysec...  */
    /* 400 / 480 line modes use 1msec here! Higher -> 0.3-0.7msec. */
    vpos2 = (i / 3125UL) + 1; /* 320ysec (conservative: 500ysec) */
    if (egamode)
      vpos2 = (i / 2500UL) + 1; /* 400ysec */
    /* sum of all vertical blanking should be something above 0.6msec  */
    /* *** If vpulse+vpos2 < 350ysec, top margin will be too small *** */

    yref1 = i * 100UL;
    yref1 /= (mode.yres + vpos1 + vpulse + vpos2);
#ifdef DMODE2
    if (yref1 > ((100UL * MAXFRAMES)+50))
      printf("Will shrink framerate from %lu to %lu Hz now...\r\n",
        (yref1+50)/100UL, MAXFRAMES);
#endif
    do {
      yref1 = i * 100UL;
      yref1 /= (mode.yres + vpos1 + vpulse + vpos2);
      if (yref1 > (100UL * MAXFRAMES)) {
        vpos2++;
      }
    } while (yref1 > (100UL * MAXFRAMES));
    printf("Found vrefresh %lu.%02lu Hz at %u + %lu + %lu + %lu.\r\n",
      yref1/100UL, yref1%100UL, mode.yres, vpos1, vpulse, vpos2);
    if (yref1 < 4950 /* 100*49.5 */) {
      printf("Framerate below 49.5 Hz is not acceptable.\r\n");
      return 1;
    }

    cparm.vrefresh = yref1; /* unit 1/100 Hz */
      /* must be 100 * clock / (htotal * vtotal) */
    cparm.clock = clock1;
    cparm.crtcflags = 0; /* no double scan, no interlace, neg/neg sync  */
      /* +h -v is classic flag for 350 lines, -h +v for 400 lines etc.! */
    if (dscan)
      cparm.crtcflags |= 1; /* activate double scan for low resolution */
      /* our timing for example describes 400 lines while 200 are visible */

#ifdef SYNCPOLARITY
    if ((mode.yres == 400) || (mode.yres == 200))
      cparm.crtcflags |= 4; /* invert hsync for 200 / 400 line modes */
    if (mode.yres == 350)
      cparm.crtcflags |= 8; /* invert vsync for EGA 350 line modes */
#endif

    pokeb(0x40,0x89, peekb(0x40,0x89) |
      ( (dscan || (mode.yres == 350)) ? 0 : 0x10 ) );
      /* bit 4 is 0 for EGA / 1 for VGA (before mode switch) */

    cparm.hsync1 = mode.xres + hpos1;
    cparm.hsync2 = cparm.hsync1 + hpulse;
    cparm.htotal = cparm.hsync2 + hpos2;

    cparm.vsync1 = mode.yres + vpos1;
    cparm.vsync2 = cparm.vsync1 + vpulse;
    cparm.vtotal = cparm.vsync2 + vpos2;

    printf("Modeline \"%ux%u\"    %lu.%02lu  %4u %4u %4u %4u  ",
      mode.xres, mode.yres, clock1/MIO, (clock1/(MIO/100UL)) % 100UL,
      mode.xres, cparm.hsync1, cparm.hsync2, cparm.htotal);
    printf("%4u %4u %4u %4u\r\n",
      mode.yres, cparm.vsync1, cparm.vsync2, cparm.vtotal);

    if (!(mode.attrib & 0x10)) { /* text mode */
      printf("Modeline \"%ux%ux%u\"  %lu.%02lu  %4u %4u %4u %4u  ",
        mode.xres, mode.yres, mode.xchar,
        clock1/MIO, (clock1/(MIO/100UL)) % 100UL,
        mode.xres, cparm.hsync1, cparm.hsync2, cparm.htotal);
      printf("%4u %4u %4u %4u  font %hux%hu\r\n",
        mode.yres, cparm.vsync1, cparm.vsync2, cparm.vtotal,
        mode.xchar, mode.ychar);
    } /* text mode */

    if ( (mode.modenumber>= 0x108) &&
         (mode.modenumber<=0x10c) &&
         (cparm.vrefresh>7200UL) ) {
      printf("OVERCLOCKING VESA text mode %x! ", mode.modenumber);
      printf("Telling BIOS that refresh would be 72.00 Hz.\r\n");
      cparm.vrefresh = 7200UL;
    }

    if (clock2) {
      unsigned int cursorpos;
#ifdef TRYVARIANTS
 foo:
#endif
      r.x.ax = 0x0300; /* get cursor pos (and size) */
      r.x.bx = 0; /* page BH */
      int86(0x10, &r, &r);
      cursorpos = r.x.dx;

      r.x.ax = 0x4f02;
      r.x.bx = mode.modenumber | 0x800 /* custom CRTC settings */
        | 0x8000 /* preserve screen contents */;
      sregs.es = FP_SEG(&(cparm.htotal));
      r.x.di = FP_OFF(&(cparm.htotal));
#if 0
//    if (peek(sregs.es, r.x.di) != cparm.htotal)
//      printf("Yikes!");
#endif
      int86x(0x10, &r, &r, &sregs);
      if (r.x.ax != 0x004f) {
        printf("Mode setting failed, code 0x%04x.\r\n", r.x.ax);
        return 1;
      }

      r.x.ax = 0x0200; /* set cursor pos */
      r.x.bx = 0; /* page BH */
      r.x.dx = cursorpos;
      int86(0x10, &r, &r);



      { /* CRTC fixup and regdump */
        int a, b;
#if CRTCFIXUP | CRTCREGDUMP
        unsigned int io = 0x3d4;
#endif

        disable();
#if CRTCFIXUP | CRTCREGDUMP
#if 0
        io = (Xinp(0x3c2) & 1) /* color? */ ? 0x3d4 : 0x3b4;
#else
        io = 0x3d4;
#endif
#endif

#ifdef CRTCFIXUP
        outportb(io, 1); /* CRTC: get hvisible */
        a = Xinp(io+1);
        outportb(io, 2); /* CRTC: get hsyncpos */
        b = Xinp(io+1);
        if (a > b) { /* hsyncpos must be <= hvisible */
          printf("FIXUP: moving hsyncpos into range.\r\n");
          outportb(io, 0x11);
          b = Xinp(io+1);
          outportb(io, 0x11);
          outportb(io+1, b & ~0x80);
          /* registers 0..7 are un-protected now */
          outportb(io, 2);
          outportb(io+1, a + 1); /* hsyncpos = hvisible+1 */
        }
#endif /* CRTCFIXUP */

#ifdef CRTCREGDUMP
        printf("CRTC[%04x]: ", io);
        for (a = 0; a < 0x18; a++) {
          outportb(io, a);
          b = Xinp(io+1);
          printf("[%02x]=%02x%s", a, b,
            ((a & 7) == 7) ? "\r\n            " : " ");
        }
        printf("\r\n");
        enable();
#endif /* CRTCREGDUMP */

      } /* CRTC fixup and regdump */

#ifdef TRYVARIANTS
      if (argc <= 4)
        return 0; /* do not enter interactive mode */

      printf("hpos1=%lu hpulse=%lu hpos2=%lu: 4 5 6 UP 1 2 3 DOWN. ESC = Abort.\r\n",
        hpos1, hpulse, hpos2);
      printf("vpos1=%lu vpulse=%lu vpos2=%lu: + * 7 UP - / 8 DOWN. ENTER = Save.\r\n<- ",
        vpos1, vpulse, vpos2);
      for (r.x.ax = 3;
           r.x.ax < (mode.xres / ((mode.xchar<8) ? 8 : mode.xchar)) - 3;
           r.x.ax++) printf("%c", '0' + (r.x.ax % 10));
      printf(" ->\r\n");

      bar: r.x.ax = 0;
      int86(0x16, &r, &r);
      r.h.cl = r.h.al; /* ASCII */

      switch (r.h.cl) {
        case 27:
          /* pokeb(0x40,0x89, peekb(0x40,0x89) | 0x10); */ /* VGA-ish */
          r.x.ax = 3; int86(0x10, &r, &r); /* standard text mode */
          return 1; /* ESC: give up */
        case 13: return 0; /* ENTER: exit and save */
        case '1': if (hpos1  >= mode.xchar) hpos1  -= mode.xchar; break;
        case '2': if (hpulse >= mode.xchar) hpulse -= mode.xchar; break;
        case '3': if (hpos2  >= mode.xchar) hpos2  -= mode.xchar; break;
        case '4': hpos1  += mode.xchar; break;
        case '5': hpulse += mode.xchar; break;
        case '6': hpos2  += mode.xchar; break;
        case '-': if (vpos1) vpos1--; break;
        case '/': if (vpulse >= 2) vpulse--; break;
        case '7': if (vpos2) vpos2--; break;
        case '+': vpos1++; break;
        case '*': vpulse++; break;
        case '8': vpos2++; break;
        default: goto bar;
      } /* switch */
    
      cparm.hsync1 = mode.xres + hpos1;
      cparm.hsync2 = cparm.hsync1 + hpulse;
      cparm.htotal = cparm.hsync2 + hpos2;

      cparm.vsync1 = mode.yres + vpos1;
      cparm.vsync2 = cparm.vsync1 + vpulse;
      cparm.vtotal = cparm.vsync2 + vpos2;

      goto foo;
#endif /* TRYVARIANTS */


    } /* unless simulating */ else {
      printf("* With VESA VBE 3.0 the new refresh rates would have been set now.\r\n");
    }

  } /* variable scope: CRTC params */

  return 0;
} /* main */

